/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.files;

import java.io.IOException;
import java.io.InputStream;
import java.io.PipedInputStream;
import java.io.PipedOutputStream;
import java.util.Objects;
import java.util.Optional;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicReference;
import java.util.function.BiFunction;
import java.util.function.Consumer;
import java.util.function.Function;
import java.util.function.Predicate;

import org.apache.commons.io.IOUtils;

import de.ozgcloud.common.errorhandling.TechnicalException;
import io.grpc.stub.StreamObserver;
import lombok.Builder;
import lombok.extern.log4j.Log4j2;

@Log4j2
class UploadServerStreamObserver<M, T> implements StreamObserver<T> {

	static final int TIMEOUT_MINUTES = 10;

	private final Consumer<T> requestVerifier;
	private final Predicate<T> hasMetadata;
	private final Function<T, M> metadataMapper;
	private final BiFunction<M, Optional<String>, FileId> metadataUploader;
	private final BiFunction<M, InputStream, CompletableFuture<String>> fileContentUploader;
	private final Function<T, byte[]> getFileContent;
	private final Consumer<FileId> responseConsumer;
	private final Consumer<String> cleanupFileContent;

	private final AtomicBoolean metadataUploaded = new AtomicBoolean(false);
	private final AtomicReference<M> fileMetadata = new AtomicReference<>();

	private PipedOutputStream pipedOutput;
	private PipedInputStream pipedInput;
	private CompletableFuture<String> contentIdFuture;

	@Builder
	UploadServerStreamObserver(Consumer<T> requestVerifier, Predicate<T> hasMetadata, Function<T, M> metadataMapper, //NOSONAR
			BiFunction<M, Optional<String>, FileId> metadataUploader, BiFunction<M, InputStream, CompletableFuture<String>> fileContentUploader,
			Function<T, byte[]> getFileContent, Consumer<FileId> responseConsumer, Consumer<String> cleanupFileContent) {
		this.requestVerifier = requestVerifier;
		this.hasMetadata = hasMetadata;
		this.metadataMapper = metadataMapper;
		this.metadataUploader = metadataUploader;
		this.fileContentUploader = fileContentUploader;
		this.getFileContent = getFileContent;
		this.responseConsumer = responseConsumer;
		this.cleanupFileContent = cleanupFileContent;
	}

	@Override
	public synchronized void onNext(T fileUploadRequest) {
		if (hasMetadata.test(fileUploadRequest)) {
			LOG.debug("Received metadata.");
			setMetadata(fileUploadRequest);
		} else {
			LOG.debug("Received file content chunk.");
			if (Objects.isNull(contentIdFuture)) {
				contentIdFuture = initContentUpload();
			}
			storeFileContent(fileUploadRequest);
		}
	}

	CompletableFuture<String> initContentUpload() {
		return fileContentUploader.apply(fileMetadata.get(), initPipedStreams());
	}

	void setMetadata(T fileUploadRequest) {
		if (metadataUploaded.getAndSet(true)) {
			throw new OzgCloudFileUploadException("Attempt to upload multiple files. Multiple uploads are not allowed.");
		}
		requestVerifier.accept(fileUploadRequest);
		fileMetadata.set(metadataMapper.apply(fileUploadRequest));
	}

	void storeFileContent(T fileUploadRequest) {
		if (Objects.isNull(fileMetadata.get())) {
			throw new OzgCloudFileUploadException("File content received before metadata.");
		}
		try {
			pipedOutput.write(getFileContent.apply(fileUploadRequest));
		} catch (IOException e) {
			throw new TechnicalException("Error when writing file content.", e);
		}
	}

	InputStream initPipedStreams() {
		try {
			pipedInput = new PipedInputStream(FileService.CHUNK_SIZE);
			pipedOutput = new PipedOutputStream(pipedInput);
			return pipedInput;
		} catch (IOException e) {
			throw new TechnicalException("Upload initialization failed", e);
		}
	}

	@Override
	public void onError(Throwable t) {
		LOG.error("Error happened. Upload stream closed", t);
		closeOutputPipe();
		closeInputPipe();
	}

	@Override
	public void onCompleted() {
		LOG.debug("Complete upload request.");
		closeOutputPipe();
		var newContentId = getContentId();
		LOG.debug("newContentId: {}", newContentId);
		try {
			var fileId = metadataUploader.apply(fileMetadata.get(), newContentId);
			responseConsumer.accept(fileId);
		} catch (RuntimeException e) {
			newContentId.ifPresent(cleanupFileContent);
			throw new TechnicalException("Cannot complete request.", e);
		}
	}

	Optional<String> getContentId() {
		try {
			return Objects.isNull(contentIdFuture) ? Optional.empty() : Optional.of(contentIdFuture.get(TIMEOUT_MINUTES, TimeUnit.MINUTES));
		} catch (ExecutionException | TimeoutException e) {
			throw new TechnicalException("Upload failed", e);
		} catch (InterruptedException e) {
			Thread.currentThread().interrupt();
			throw new TechnicalException("Upload was interrupted", e);
		} finally {
			closeInputPipe();
		}
	}

	void closeOutputPipe() {
		IOUtils.closeQuietly(pipedOutput, e -> LOG.error("Cannot close output stream", e));
	}

	void closeInputPipe() {
		IOUtils.closeQuietly(pipedInput, e -> LOG.error("Cannot close input stream", e));
	}
}